import pynwb
import os
from orgutils import orgutils
from datetime import datetime as dt
import SAPOFTO


def createSubject(args, subject_id=None):
    # survey user about subject information
    #dotnwborglab = SAPOFTO.SAPOFTO(key='dotnwborglab', filename='.nwborglab.org')

    #survey_node = dotnwborglab['SURVEYS']['SUBJECT CREATE']

    #for question_key in survey_node.key():
        

    
    if not os.path.isfile('.nwborglab.org'):
        print('NWBORG ERROR: not in the uppermost folder of an nwborg project.')
        return


    
    print('Create Subject!')
    if subject_id == None:
        subject_id = input('Please input the subject\'s ID, or press enter to autogenerate a new ID')
    if subject_id == '':
        subject_files = os.listdir('subjects')
        largest_id = 0
        for subject_file in subject_files:
            subject_num = int(subject_file)
            if subject_num > largest_id:
                largest_id = subject_num
        new_id = largest_id + 1
        preceding_zeroes = ''
        if new_id < 1000:
            preceding_zeroes += '0'
        if new_id < 100:
            preceding_zeroes += '0'
        if new_id < 10:
            preceding_zeroes += '0'
        subject_id = preceding_zeroes + str(new_id)
    
    print('Proceeding with id ' + subject_id + ' for subject creation...')
    
    subject_dict = dict()
    subject_dict['Subject ID'] = subject_id
    subject_dict['Species'] = input('Please enter subject species...')
    subject_dict['Age'] = input('Please enter subject age in any desired units (ex. "23 years")...')
    subject_dict['Sex'] = input('Please enter subject sex M for male F for female O for other...')
    subject_dict['Date of Birth'] = input('Please input the subject\'s date of birth (ex. 08/11/1998)...')
    subject_dict['Description'] = input('Please provide additional description for the subject...')
    os.mkdir('subjects/' + subject_id)
    orgutils.dictToOrg(org_data=subject_dict,output_filename='subjects/' + subject_id + '/' + subject_id + '.org')
    print('subject info saved to subjects/' + subject_id + '.org')
    



def subjectIdToNwbSubjectObject(subject_id):
    subject_node = SAPOFTO.SAPOFTO(os.path.join('Subjects',str(subject_id),str(subject_id),'.org'))

    subject_nwb = pynwb.file.Subject()
    if 'AGE' in subject_node.keys():
        subject_nwb.age = subject_node['AGE'].getValue()
    if 'DESCRIPTION' in subject_node.keys():
        subject_nwb.description = subject_node['DESCRIPTION'].getValue()
    if 'GENOTYPE' in subject_node.keys():
        subject_nwb.genotype = subject_node['GENOTYPE'].getValue()
    if 'SEX' in subject_node.keys():
        subject_nwb.sex = subject_node['SEX'].getValue()
    if 'SPECIES' in subject_node.keys():
        subject_nwb.species = subject_node['SPECIES'].getValue()
    if 'SUBJECT_ID' in subject_node.keys():
        subject_nwb.subject_id = subject_node['SUBJECT_ID'].getValue()
    if 'WEIGHT' in subject_node.keys():
        subject_nwb.weight = subject_node['WEIGHT'].getValue()
    if 'STRAIN' in subject_node.keys():
        subject_nwb.strain = subject_node['STRAIN'].getValue()
    if 'DATE_OF_BIRTH' in subject_node.keys():
        subject_nwb.date_of_birth = subject_node['DATE_OF_BIRTH'].getValue()
    return subject_nwb



import nwborglab
import os
from os.path import expanduser
from datetime import datetime as dt
from SAPOFTO import SAPOFTO, recursiveFolderWrite, recursiveFolderRead
from pynwb import NWBHDF5IO



def init(args):
    # print ascii art
    # looks messed up in .py src cause you have to escape the escapes
    # aka M-x replace-string \ --> \\ -->
    print("""
      ___           ___           ___     
     /\\__\\         /\\__\\         /\\  \\    
    /::|  |       /:/ _/_       /::\\  \\   
   /:|:|  |      /:/ /\\__\\     /:/\\:\\  \\  
  /:/|:|  |__   /:/ /:/ _/_   /::\\~\\:\\__\\ 
 /:/ |:| /\\__\\ /:/_/:/ /\\__\\ /:/\\:\\ \\:|__|
 \\/__|:|/:/  / \\:\\/:/ /:/  / \\:\\~\\:\\/:/  /
     |:/:/  /   \\::/_/:/  /   \\:\\ \\::/  / 
     |::/  /     \\:\\/:/  /     \\:\\/:/  /  
     /:/  /       \\::/  /       \\::/__/   
     \\/__/         \\/__/         ~~       
      ___           ___           ___     
     /\\  \\         /\\  \\         /\\  \\    
    /::\\  \\       /::\\  \\       /::\\  \\   
   /:/\\:\\  \\     /:/\\:\\  \\     /:/\\:\\  \\  
  /:/  \\:\\  \\   /::\\~\\:\\  \\   /:/  \\:\\  \\ 
 /:/__/ \\:\\__\\ /:/\\:\\ \\:\\__\\ /:/__/_\\:\\__\\
 \\:\\  \\ /:/  / \\/_|::\\/:/  / \\:\\  /\\ \\/__/
  \\:\\  /:/  /     |:|::/  /   \\:\\ \\:\\__\\  
   \\:\\/:/  /      |:|\\/__/     \\:\\/:/  /  
    \\::/  /       |:|  |        \\::/  /   
     \\/__/         \\|__|         \\/__/
      ___        ___            ___     
     /\\__\\      /\\  \\          /\\  \\    
    /:/  /     /::\\  \\        /::\\  \\   
   /:/  /     /:/\\:\\  \\      /:/\\:\\  \\  
  /:/  /     /::\\~\\:\\  \\    /::\\~\\:\\__\\ 
 /:/__/     /:/\\:\\ \\:\\__\\  /:/\\:\\ \\:|__|
 \\:\\  \\     \\/__\\:\\/:/   / \\:\\~\\:\\/:/  /
  \\:\\  \\         \\::/  /    \\:\\ \\::/  / 
   \\:\\  \\        /:/  /      \\:\\/:/  /  
    \\:\\__\\      /:/  /        \\::/__/   
     \\/__/      \\/__/          ~~       


       		   		   v0.0.1""")
    
    if len(args) > 0:
        print('attempting nwborglab initialization from file')
        if not os.isfile(args[0]) :
            print(args[0] + ' no such file :(')
        if not args[0].endswith('.nwb'):
            io = NWBHDF5IO(args[0], 'r')

        print(nwbfile_in.get_scratch('nwborglab_sensors'))
        sensors_sapofto = SAPOFTO('sensors',nwbfile_in.get_scratch('nwborglab_sensors'), case_sensitive=True)
        overview_sapofto = SAPOFTO('OVERVIEW',nwbfile_in.get_scratch('nwborglab_overview'))
        session_skeleton_sapofto = nwbfile_in.get_scratch('nwborglab_session_skeleton')
        session_skeleton_sapofto = nwbfile_in.get_scratch('nwborglab_session_source_code')
        
        recursiveFolderWrite(os.getcwd())
        return
    
    tab = '    '
    half_tab = '  '

    home = expanduser("~")
    if not os.path.isfile(os.path.join(home,'.nwborglabsrc.org')):
        print('First nwborglab initialization! Setting up... ')
        with open('/'.join(nwborglab.__file__.split('/')[:-1]) + '/' + '.nwborglabsrc.org', 'r') as f: # TODO replace this with where the package will be
            content = f.read()

        with open(os.path.join(home,'.nwborglabsrc.org'), 'w') as f: # TODO replace this with where the package will be
            f.write(content)

    if os.path.isfile('.nwborglab.org'):
        print('ERROR: This directory is already the root of an nwborg project.')
        return

    # TODO fill out the .nwborg
    init_org = SAPOFTO.SAPOFTO('.nwborglab')
    nwborglabsrc_sapofto = SAPOFTO.SAPOFTO(key='NWBORGLABSRC',filename=os.path.join(home,'.nwborglabsrc.org'))

    with open('overview.org', 'w') as f:
        f.write(str(nwborglabsrc_sapofto['OVERVIEW SRC'].getValue()))
        nwborglabsrc_sapofto.pop('OVERVIEW SRC')
    with open('session_skeletons.org', 'w') as f:
        f.write(str(nwborglabsrc_sapofto['SESSION_SKELETONS SRC'].getValue()))
        nwborglabsrc_sapofto.pop('SESSION_SKELETONS SRC')
    with open('.nwborglab.org', 'w') as f:
        f.write(str(nwborglabsrc_sapofto.castOrgLiteral()))
    # TODO make .nwb.org in each subdirectory
    # and make root path a variable accessible in the project no matter where you are at
    # make directories if necessary
    if not os.path.isdir('subjects'):
        print('creating subjects directory...')
        os.mkdir('subjects')
        print('complete.')
    else:
        print('subjects directory found')

    if not os.path.isdir('sensors'):
        print('creating sensors directory...')
        os.mkdir('sensors')
        print('complete.')
    else:        
        print('sensors directory found')
        
    if not os.path.isdir('stimuli'):
        print('creating stimuli directory...')
        os.mkdir('stimuli')
        print('complete.')
    else:
        print('stimuli directory found')
        
    if not os.path.isdir('sessions'):
        print('creating sessions directory...')
        os.mkdir('sessions')
        print('complete.')
    else:
        print('sessions directory found')

    if not os.path.isdir('src'):
        print('creating src directory...')
        os.mkdir('src')
        print('complete.')
    else:
        print('src directory found')



from nwborglab import subjectHandler
import os
from SAPOFTO import SAPOFTO
import orgutils
from datetime import datetime as dt


def scheduleSession(args):
    if not os.path.isfile('.nwborglab.org'):
        print('NWBORG ERROR: not in the uppermost folder of an nwborg project.')
        return
    
    session_skeletons = SAPOFTO.SAPOFTO(key='session_skeletons',filename='session_skeletons.org')
    print('args:', args)
    session_archetype = args[0]
    skeleton = session_skeletons[session_archetype]
    for role in skeleton['subject roles'].keys():
        subject_id = input('Please enter a subject id to fill the role of ' + str(role) + ' in this session...')
        if not os.path.isdir(os.path.join('subjects', str(subject_id))):
            print('Subject ' + subject_id + ' not found, proceeding to subject registration...')
            subjectHandler.createSubject(args,subject_id=subject_id)                    
        print('Proceeding with subject id ' + subject_id + ' filling the roll of ' + role + '...')
        # ACTUAL TODO error handling for input here
        skeleton['SUBJECT ROLES'][role.upper()].addChild(SAPOFTO.SAPOFTO('SUBJECT ID',content=str(subject_id) ))
        
        
    if not os.path.isdir('sessions/' + session_archetype):
        print('creating sessions directory...')
        os.mkdir('sessions/' + session_archetype)
        print('complete.')
    
    session_id = input('Please input the session\'s ID, or press enter to autogenerate a new ID...')
    if session_id == '':
        session_files = os.listdir('sessions/' + session_archetype)
        largest_id = 0
        for session_file in session_files:
            if '.' in session_file:
                continue
            session_num = int(session_file)
            if session_num > largest_id:
                largest_id = session_num
        new_id = largest_id + 1
        preceding_zeroes = ''
        if new_id < 1000:
            preceding_zeroes += '0'
        if new_id < 100:
            preceding_zeroes += '0'
        if new_id < 10:
            preceding_zeroes += '0'
        session_id = preceding_zeroes + str(new_id)

    if not os.path.isdir('sessions/' + session_archetype + '/' + session_id):
        print('creating directory for session ' + session_id + '...')
        os.mkdir('sessions/' + session_archetype + '/' + session_id)
        print('complete.')

    print('Proceeding with id ' + session_id + ' for session creation...')
    session_dict = dict()
    session_dict['subject roles'] = skeleton['subject roles']
    session_dict['date'] = input('Please input the date expected for this session to be recorded(ex. 08/11/1998)...')
    session_dict['session supervisor'] = input('Please input the name of the experimenter responsible for this session being recorded...')
    orgutils.dictToOrg(org_data=session_dict,output_filename='sessions/' + session_archetype + '/' + session_id + '/session.org')
    print('session info saved to sessions/' + session_archetype + '/' + session_id + '/' + 'session.org')
    # Read in the hardware configs for the sensors pertaining to the different roles
    hardware_used = dict()
    sensor_files = set()
    for role in session_dict['subject roles'].keys():
        hardware_used[role] = session_dict['subject roles'][role]['sensors']
        sensor_files.add(session_dict['subject roles'][role]['sensors']) # TODO add .split functionality for lists of sensors
        # ^^^^ Also you can probably just keep track of this better by seeing where orgutils is opening org files.
        # ^^^^ Better yet maybe orgutils feature incoming?
        




def generateSessionCode(args):
    session_skeletons = SAPOFTO.SAPOFTO(key='session_skeletons',filename='session_skeletons.org')
    print('args:', args)
    session_archetype = args[0]
    skeleton = session_skeletons[session_archetype]

    session_sapofto = SAPOFTO.SAPOFTO('SESSION SAPOFTO')
    print(session_sapofto.keys())

    session_sapofto.addChild(skeleton['SUBJECT ROLES'])

    # Read in the hardware configs for the sensors pertaining to the different roles
    hardware_used = SAPOFTO.SAPOFTO('HARDWARE')
    sensor_files = set()
    for role in session_sapofto['SUBJECT ROLES'].keys():
        hardware_used.addChild(SAPOFTO.SAPOFTO(role, ''))
        hardware_used[role].append(session_sapofto['SUBJECT ROLES'][role.upper()]['SENSORS'].getValue())
        sensor_files.add(session_sapofto['SUBJECT ROLES'][role.upper()]['SENSORS'].getValue()) # TODO add .split functionality for lists of sensors
        
        # ^^^^ Also you can probably just keep track of this better by seeing where orgutils is opening org files.
        # ^^^^ Better yet maybe orgutils feature incoming?


    # source code generation 
    print(session_archetype + '/run.py not found... generating now....')
    tab = '    '
    half_tab = '  '

    # should be ubiquitous to all sessions, probably should be an alternate command acutally
    # that happens to run if run.py doesn't exist
    if not os.path.isdir('sessions/' + session_archetype):
        print('creating sessions directory...')
        os.mkdir('sessions/' + session_archetype)
        print('complete.')

    session_code = SAPOFTO.SAPOFTO('SESSION CODE')
    # autogenerated nwb initializations
    # imports
    dotnwborglab = SAPOFTO.SAPOFTO(key='dotnwborglab', filename='.nwborglab.org')

    # contstruct prototype parameters
    prototype_parameters = SAPOFTO.SAPOFTO(key='prototype_parameters')
    prototype_parameters.addChild(SAPOFTO.SAPOFTO('SESSION_ARCHETYPE', content=session_archetype))

    dotnwborglab.populatePrototype(prototype_parameters)
    session_code.appendLine("#!!! This file was automatically generated by nwborg !!!")
    session_code.appendLine("# Code generated primarily from the parsing of: ")
    session_code.appendLine("#   - session_skeletons.org")
    for sensor in sensor_files:
        session_code.appendLine("#   - sensors/" + str(sensor) + ".org")

    session_code.append(('\n').join(dotnwborglab['SESSION IMPORTS'].lineList(with_tab=True, tab_displacement=-1)))
    session_code.appendLine('')
    #header_proto = SAPOFTO.SAPOFTO('SESSION_ARCHETYPE', content=session_archetype)
    dotnwborglab['SESSION HEADER'].populatePrototype(prototype_parameters)
    session_code.append(dotnwborglab['SESSION HEADER'].getValue())
    #print(skeleton['nwb unit columns'])
    if 'NWB UNIT COLUMNS' in skeleton.keys():
        for unit_column in skeleton['NWB UNIT COLUMNS'].keys():
            session_code.appendLine('\n'+tab+"nwbfile.add_unit_column('"+str(unit_column)+"','"+skeleton['nwb unit columns'][unit_column].getValue().replace('\n', ' ')+"')")

    # ACTUAL TODO add argument for session id for run.py and have...
    # ...it print the session data as a confirmation

    # user defined additions to the source code
    # one at a time add the sensors and add prompts to the file ensuring

    # SURVEY 
    for role in skeleton['SUBJECT ROLES'].keys():
        if 'SURVEY' in skeleton['SUBJECT ROLES'][role].keys():
            session_code.appendLine('\n'+tab+'print("Survey for ' + role + ':")')
            session_code.appendLine('\n'+tab+'skeleton["SUBJECT ROLES"]["'+role+'"]["SURVEY"]["INITIAL"].applySurvey()')
                        
    session_code.append('\n'+tab + ('\n'+tab).join(skeleton['code']['initial'].lineList(with_tab=True, tab_displacement=-1)))

    for role in hardware_used.keys():
        print('initializing sensors for the ' + role + '...')
        sensor = hardware_used[role].getValue()
        config = SAPOFTO.SAPOFTO(key=sensor + '_config', filename='sensors/' + sensor + '.org')
        #for sensor in hardware_used[role]:

        initialization_code = config['code']['initial'].lineList(with_tab=True, tab_displacement=-1)
        session_code.append('\n' + tab + ('\n'+tab).join(initialization_code))

    session_code.appendLine('\n' + tab + 'try:')
    session_code.appendLine((tab * 2) + 'while(True):')

    for role in hardware_used.keys():
        print('initializing sensors for the ' + role + '...')
        #for sensor in hardware_used[role]:
        sensor = hardware_used[role].getValue()
        config = SAPOFTO.SAPOFTO(key=sensor + '_config',filename='sensors/' + sensor + '.org')
        initialization_code = config['code']['loop'].lineList(with_tab=True, tab_displacement=-1)
        session_code.append(('\n'+(tab*3)) + ('\n'+(tab*3)).join(initialization_code))

    session_code.append('\n'+tab + ('\n'+(tab*3)).join(skeleton['code']['loop'].lineList(with_tab=True, tab_displacement=-1)))
    session_code.appendLine('\n' + tab + 'except:')
    
    session_code.appendLine('\n' + tab * 2 + 'print("recording complete")')#, e)')
    # Apply survey questions
    #

    
    # ACTUAL TODO FIX THE DEBUG ^^^^^^
    # ACTUAL TODO FIX THE OUTPUT vvvvv

    for role in hardware_used.keys():
        print('initializing sensors for the "' + role + '" role...')
        #for sensor in hardware_used[role]:
        sensor = hardware_used[role].getValue()
        config = SAPOFTO.SAPOFTO(key=sensor + '_config',filename='sensors/' + sensor + '.org')
        initialization_code = config['code']['terminal'].lineList(with_tab=True, tab_displacement=-1)
        session_code.append(('\n'+(tab*2)) + ('\n'+(tab*2)).join(initialization_code))

    for role in skeleton['SUBJECT ROLES'].keys():
        if 'SURVEY' in skeleton['SUBJECT ROLES'][role].keys():
            session_code.appendLine('\n'+tab*2+'print("Survey for ' + role + ':")')
            session_code.appendLine('\n'+tab*2+'skeleton["SUBJECT ROLES"]["'+role+'"]["SURVEY"]["TERMINAL"].applySurvey()')



    session_code.append('\n'+tab*2 + ('\n'+(tab*2)).join(skeleton['code']['terminal'].lineList(with_tab=True, tab_displacement=-1)))
    dotnwborglab['SESSION FOOTER'].populatePrototype(prototype_parameters)
    session_code.append('\n'+tab+dotnwborglab['SESSION FOOTER'].getValue())
    session_code.appendLine('\n'+tab+"with NWBHDF5IO(str(session_id) + '.nwb', 'w') as io:")
    session_code.appendLine(tab*2+"io.write(nwbfile)")

    # vvvvvvvv This should be the last step in the code generation process vvvvvvvv
    # Make callable from terminal and add call to main
    session_code.appendLine("\nif __name__ == '__main__':")
    session_code.appendLine(tab + "main()")
    session_code.writeToFile(os.path.join('sessions', session_archetype + '/'), 'run.py' )





def quickstartSession(args, session_dict=None, subject_id_dict=None, session_id=None, unknown=[], session_sapofto=None):
    # PICK IT UP 998
    if not os.path.isfile('.nwborglab.org'):
        print('NWBORG ERROR: not in the uppermost folder of an nwborg project.')
        return
    
    session_skeletons = SAPOFTO.SAPOFTO(key='SKELETON',filename='session_skeletons.org')
    print('args:', args)
    session_archetype = args[0]
    skeleton = session_skeletons[session_archetype]
    for role in skeleton['subject roles'].keys():
        if 'SUBJECT ID' in skeleton['SUBJECT ROLES'][role.upper()].keys():            
            subject_id = skeleton['SUBJECT ROLES'][role.upper()]['SUBJECT ID'].getValue()            
            #ACTUAL TODO make sure its valid
        if subject_id_dict != None and role in subject_id_dict.keys():
            # subject id specified via passing in the argument
            subject_id = subject_id_dict[role]['SUBJECT ID'].getValue()
        else:
            subject_id = input('Please enter a subject id to fill the role of ' + str(role) + ' in this session...')
        print(subject_id)
        print('Proceeding with subject id ' + subject_id + ' filling the roll of ' + role + '...')
        # ACTUAL TODO error handling for input here
        #skeleton['SUBJECT ROLES'][role]['SUBJECT ID'] = subject_id
        skeleton['SUBJECT ROLES'][role.upper()].addChild(SAPOFTO.SAPOFTO('SUBJECT ID',content=str(subject_id) ))
                
    if not os.path.isdir('sessions/' + session_archetype):
        print('creating sessions directory...')
        os.mkdir('sessions/' + session_archetype)
        print('complete.')
    if session_id == None:  # i.e. if session_id was not passed as a parameter to this function
        session_id = input('Please input the session\'s ID, or press enter to autogenerate a new ID...')
    if session_id == '':
        session_files = os.listdir('sessions/' + session_archetype)
        largest_id = 0
        for session_file in session_files:
            if '.' in session_file:
                continue
            session_num = int(session_file)
            if session_num > largest_id:
                largest_id = session_num
        new_id = largest_id + 1
        preceding_zeroes = ''
        if new_id < 1000:
            preceding_zeroes += '0'
        if new_id < 100:
            preceding_zeroes += '0'
        if new_id < 10:
            preceding_zeroes += '0'
        session_id = preceding_zeroes + str(new_id)

    if not os.path.isdir('sessions/' + session_archetype + '/' + session_id):
        print('creating directory for session ' + session_id + '...')
        os.mkdir('sessions/' + session_archetype + '/' + session_id)
        print('complete.')

    print('Proceeding with id ' + session_id + ' for session creation...')
    if session_sapofto == None: # i.e. if session_sapofto was not passed as a parameter to this function
        session_sapofto = SAPOFTO.SAPOFTO('SESSION METADATA')
    session_sapofto.addChild(skeleton['SUBJECT ROLES'])
    session_sapofto.addChild(SAPOFTO.SAPOFTO('ARCHETYPE', content=session_archetype))
    if not 'DATE' in session_sapofto.keys():
        session_sapofto.addChild(SAPOFTO.SAPOFTO('DATE', content=str(dt.today())))
    
    if not 'SESSION SUPERVISOR' in session_sapofto.keys():
        session_sapofto.addChild(SAPOFTO.SAPOFTO('SESSION SUPERVISOR', input('Please input the name of the experimenter responsible for this session being recorded...')))
    print(session_sapofto)
    session_sapofto.writeToFile('sessions/' + session_archetype + '/' + session_id + '/', 'session.org')

    print('session info saved to sessions/' + session_archetype + '/' + session_id + '/' + 'session.org')

    if not os.path.isfile('sessions/' + session_archetype + '/run.py'):
        generateSessionCode(args)

    # start session
    print('\npython sessions/' + session_archetype + '/run.py --session-id ' + session_id +' '+ ' '.join(unknown) + '\n')
    os.system('python sessions/' + session_archetype + '/run.py --session-id ' + session_id +' '+ ' '.join(unknown))
    return {'args' : args, 'session_sapofto' : session_sapofto, 'subject_id_dict' : session_sapofto['subject roles']}
